import numpy
import unittest
import itertools
from .. import datamerging
from ...utils.math import asproduct


class test_data_merging(unittest.TestCase):

    def test_general(self):
        shape = 3, 7
        dtype = float
        data = numpy.arange(numpy.product(shape), dtype=dtype)
        data = data.reshape(shape)
        axis = 0, 1, -1
        newaxis = True, False
        reshape = False, True
        order = 'C', 'F'
        advanced = False, True

        options = itertools.product(axis, newaxis, reshape, order, advanced)
        for axis, newaxis, reshape, order, advanced in options:
            kwargs = {'axis': axis,
                      'newaxis': newaxis,
                      'order': order,
                      'allow_advanced_indexing': advanced}
            with self.subTest(reshape=reshape, **kwargs):
                sources = [data]*3
                shapes = [shape]*3
                if not newaxis:
                    extrashape = list(shape)
                    extrashape[axis] *= 2
                    extrashape = tuple(extrashape)
                    extradata = numpy.arange(numpy.product(extrashape), dtype=dtype)
                    extradata = extradata.reshape(extrashape)
                    sources.append(extradata)
                    shapes.append(extrashape)
                if newaxis:
                    edata = numpy.stack(sources, axis=axis)
                else:
                    edata = numpy.concatenate(sources, axis=axis)
                if reshape:
                    dshape = edata.shape
                    it = asproduct(numpy.product(dshape), 3, includeone=False)
                    nshape = next(it)
                    while nshape == dshape:
                        nshape = next(it)
                    edata = edata.reshape(nshape, order=order)
                else:
                    nshape = None
                mshape, fillgen = datamerging.mergeGenerator(sources, shapes,
                                                             shape=nshape,
                                                             **kwargs)
                mdata = numpy.empty(shape=mshape, dtype=dtype)
                for ndarray, idx_generator in fillgen():
                        for idxin, idxout in idx_generator():
                            for idx in idxin, idxout:
                                isadvanced = any(isinstance(x, numpy.ndarray) for x in idx)
                                self.assertEqual(isadvanced, advanced and reshape)
                            mdata[idxout] = ndarray[idxin]
                numpy.testing.assert_array_equal(edata, mdata)

    def test_single(self):
        order = 'C', 'F'
        advanced = False, True
        options = itertools.product(order, advanced)
        dshape = 13, 14
        scanshape1 = 6, 2, 3
        scanshape2 = numpy.product(scanshape1, dtype=int),
        for order, advanced in options:
            if order == 'C':
                shape1 = scanshape1 + dshape
                shape2 = scanshape2 + dshape
            else:
                shape1 = dshape + scanshape1
                shape2 = dshape + scanshape2
            n = numpy.product(shape1, dtype=int)
            data = numpy.arange(n, dtype=float).reshape(shape1)
            self._assert_single(data, shape2, order, advanced)
            data = numpy.arange(n, dtype=float).reshape(shape2)
            self._assert_single(data, shape1, order, advanced)

    def _assert_single(self, data, shape, order, advanced):
        edata = data.reshape(shape, order=order)
        mshape, fillgen = datamerging.mergeGenerator([data], [data.shape],
                                                     shape=shape,
                                                     newaxis=False,
                                                     order=order,
                                                     allow_advanced_indexing=advanced)

        self.assertEqual(shape, mshape)
        mdata = numpy.empty(shape=mshape, dtype=data.dtype)
        for ndarray, idx_generator in fillgen():
            for idxin, idxout in idx_generator():
                mdata[idxout] = ndarray[idxin]
        numpy.testing.assert_array_equal(edata, mdata)

    def test_intlist_to_slice(self):
        idx = (numpy.array([4, 0, 1, 2, 3, 4, 0]),
               numpy.array([2, 3, 3, 3, 3, 3, 4]))
        self._assert_intlist_to_slice(idx)
        idx = (numpy.array([2, 1, 0, 1, 2, 3, 4, 0]),
               numpy.array([2, 2, 2, 3, 3, 3, 3, 4]))
        self._assert_intlist_to_slice(idx)
        idx = (numpy.array([0,0,0,0,1,1,1,1,2,2,2,2,6,4,2,2,2]),
               numpy.array([0,1,2,3,0,1,2,3,3,3,3,3,3,3,3,3,3]),
               numpy.array([0,1,0,1,0,1,2,3,4,5,6,7,1,2,3,9,8]))
        self._assert_intlist_to_slice(idx)
        idx = (numpy.array([0]),
               numpy.array([0]),
               numpy.array([0]))
        self._assert_intlist_to_slice(idx)
        idx = (numpy.array([0, 0]),
               numpy.array([0, 2]),
               numpy.array([0, 4]))
        self._assert_intlist_to_slice(idx)
        idx = (numpy.array([0, 0, 1]),
               numpy.array([0, 2, 2]),
               numpy.array([0, 4, 4]))
        self._assert_intlist_to_slice(idx)
        idx = (numpy.array([1, 0, 0]),
               numpy.array([2, 0, 2]),
               numpy.array([4, 0, 4]))
        self._assert_intlist_to_slice(idx)
        
        combinations = []
        comb = [10], [20], [30]
        combinations.append(numpy.array(comb))
        comb = [1], [2], [3]
        combinations.append(numpy.array(comb))
        comb = [0]*8, list(range(8)), list(range(1, 9))
        combinations.append(numpy.array(comb))
        comb = list(range(4))+list(range(4)), [0]*8, list(range(3))+list(range(5))
        combinations.append(numpy.array(comb))
        for comb in itertools.permutations(combinations):
            idx = numpy.concatenate(comb, axis=1)
            self._assert_intlist_to_slice(idx)

    def _assert_intlist_to_slice(self, idx):
        for slicetwo in True, False:
            for allow_negative_stride in True, False:
                idx = numpy.asarray(idx)
                x = numpy.arange(idx.max()+1)
                y = []
                for idx2 in datamerging.intListToSliceIndexing(idx, slicetwo=slicetwo,
                                                               allow_negative_stride=allow_negative_stride):
                    if not allow_negative_stride:
                        self.assertTrue(all(idx3.step>0 for idx3 in idx2
                                            if isinstance(idx3, slice)))
                    lst = [numpy.array(x[idx3]) for idx3 in idx2]
                    n = max(arr.size for arr in lst)
                    lst = [arr if arr.ndim else arr.repeat(n)
                           for arr in lst]
                    y.append(numpy.array(lst))
                idx2 = numpy.concatenate(y, axis=1)
                numpy.testing.assert_array_equal(idx, idx2)


def test_suite():
    """Test suite including all test suites"""
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_data_merging("test_general"))
    testSuite.addTest(test_data_merging("test_single"))
    testSuite.addTest(test_data_merging("test_intlist_to_slice"))
    return testSuite


if __name__ == '__main__':
    import sys

    mysuite = test_suite()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)
